use bitflags::bitflags;

pub const PPU_REG_ADDR_CTRL: u16 = 0x2000;
pub const PPU_REG_ADDR_MASK: u16 = 0x2001;
pub const PPU_REG_ADDR_STATUS: u16 = 0x2002;
pub const PPU_REG_ADDR_OAMADDR: u16 = 0x2003;
pub const PPU_REG_ADDR_OAMDATA: u16 = 0x2004;
pub const PPU_REG_ADDR_SCROLL: u16 = 0x2005;
pub const PPU_REG_ADDR_ADDR: u16 = 0x2006;
pub const PPU_REG_ADDR_DATA: u16 = 0x2007;
pub const PPU_REG_ADDR_OAMDMA: u16 = 0x4014;

pub struct PpuRegisters {
    /// PPUCTRL ($2000, W) -
    /// NMI enable (V), PPU master/slave (P), sprite height (H), background tile select (B), sprite tile select (S), increment mode (I), nametable select / X and Y scroll bit 8 (NN)
    //  7  bit  0
    //  ---- ----
    //  VPHB SINN
    //  |||| ||||
    //  |||| ||++- Base nametable address
    //  |||| ||    (0 = $2000; 1 = $2400; 2 = $2800; 3 = $2C00)
    //  |||| |+--- VRAM address increment per CPU read/write of PPUDATA
    //  |||| |     (0: add 1, going across; 1: add 32, going down)
    //  |||| +---- Sprite pattern table address for 8x8 sprites
    //  ||||       (0: $0000; 1: $1000; ignored in 8x16 mode)
    //  |||+------ Background pattern table address (0: $0000; 1: $1000)
    //  ||+------- Sprite size (0: 8x8 pixels; 1: 8x16 pixels – see PPU OAM#Byte 1)
    //  |+-------- PPU master/slave select
    //  |          (0: read backdrop from EXT pins; 1: output color on EXT pins)
    //  +--------- Vblank NMI enable (0: off, 1: on)
    pub control: u8,

    /// PPUMASK ($2001, W) -
    /// Color emphasis (BGR), sprite enable (s), background enable (b), sprite left column enable (M), background left column enable (m), greyscale (G)
    //  7  bit  0
    //  ---- ----
    //  BGRs bMmG
    //  |||| ||||
    //  |||| |||+- Greyscale (0: normal color, 1: greyscale)
    //  |||| ||+-- 1: Show background in leftmost 8 pixels of screen, 0: Hide
    //  |||| |+--- 1: Show sprites in leftmost 8 pixels of screen, 0: Hide
    //  |||| +---- 1: Enable background rendering
    //  |||+------ 1: Enable sprite rendering
    //  ||+------- Emphasize red (green on PAL/Dendy)
    //  |+-------- Emphasize green (red on PAL/Dendy)
    //  +--------- Emphasize blue
    pub mask: u8,

    /// PPUSTATUS ($2002, R) -
    /// Vblank (V), sprite 0 hit (S), sprite overflow (O); read resets write pair for $2005/$2006
    //  7  bit  0
    //  ---- ----
    //  VSOx xxxx
    //  |||| ||||
    //  |||+-++++- (PPU open bus or 2C05 PPU identifier)
    //  ||+------- Sprite overflow flag
    //  |+-------- Sprite 0 hit flag
    //  +--------- Vblank flag, cleared on read. Unreliable; see below.
    pub status: u8,

    /// OAMADDR ($2003, W) - OAM read/write address
    //  7  bit  0
    //  ---- ----
    //  AAAA AAAA
    //  |||| ||||
    //  ++++-++++- OAM address
    pub oam_addr: u8,

    /// OAMDATA ($2004, RW) - OAM data read/write
    //  7  bit  0
    //  ---- ----
    //  DDDD DDDD
    //  |||| ||||
    //  ++++-++++- OAM data
    pub oam_data: u8,

    /// PPUSCROLL ($2005, Wx2) - X and Y scroll bits 7-0 (two writes: X scroll, then Y scroll)
    //  1st write
    //  7  bit  0
    //  ---- ----
    //  XXXX XXXX
    //  |||| ||||
    //  ++++-++++- X scroll bits 7-0 (bit 8 in PPUCTRL bit 0)
    //
    //  2nd write
    //  7  bit  0
    //  ---- ----
    //  YYYY YYYY
    //  |||| ||||
    //  ++++-++++- Y scroll bits 7-0 (bit 8 in PPUCTRL bit 1)
    pub scroll: u8,

    /// PPUADDR ($2006, Wx2) - VRAM address (two writes: most significant byte, then least significant byte)
    //  1st write  2nd write
    //  15 bit  8  7  bit  0
    //  ---- ----  ---- ----
    //  ..AA AAAA  AAAA AAAA
    //    || ||||  |||| ||||
    //    ++-++++--++++-++++- VRAM address
    pub addr: u8,

    /// PPUDATA ($2007, RW) - VRAM data read/write
    //  7  bit  0
    //  ---- ----
    //  DDDD DDDD
    //  |||| ||||
    //  ++++-++++- VRAM data
    pub data: u8,

    /// OAMDMA ($4014, W) - OAM DMA high address
    //  7  bit  0
    //  ---- ----
    //  AAAA AAAA
    //  |||| ||||
    //  ++++-++++- Source page (high byte of source address)
    pub oam_dma: u8
}

pub struct PpuInternalRegisters {

    /// During rendering, used for the scroll position. Outside of rendering, used as the current VRAM address.
    pub v: u16,

    /// During rendering, specifies the starting coarse-x scroll for the next scanline and the starting y scroll for the screen.
    /// Outside of rendering, holds the scroll or VRAM address before transferring it to v.
    pub t: u16,

    /// The fine-x position of the current scroll, used during rendering alongside v.
    pub x: u8,

    /// Toggles on each write to either PPUSCROLL or PPUADDR, indicating whether this is the first or second write. Clears on reads of PPUSTATUS.
    /// Sometimes called the 'write latch' or 'write toggle'.
    pub w: bool,
}

bitflags! {
    pub struct PpuControl: u8 {
        const BaseNametableAddress = 0b0000_0011;
        const VramAdressIncrement = 0b0000_0100;
        const SpritePatternTableAddress = 0b0000_1000;
        const BackgroundPatternTableAddress = 0b0001_0000;
        const SpriteSize = 0b0010_0000;
        const PpuMasterSlave = 0b0100_0000;
        const VBlankNmiEnable = 0b1000_0000;
    }
}

bitflags! {
    pub struct PpuMask: u8 {
        const Greyscale = 0b0000_0001;
        const ShowLeftBackground = 0b0000_0010;
        const ShowLeftSprites = 0b0000_0100;
        const EnableBackgroundRendering = 0b0000_1000;
        const EnableSpriteRendering = 0b0001_0000;
        const EmphasizeRed = 0b0010_0000;
        const EmphasizeGreen = 0b0100_0000;
        const EmphasizeBlue = 0b1000_0000;
    }
}

bitflags! {
    pub struct PpuStatus: u8 {
        const OpenBus = 0b0001_1111;
        const SpriteOverflow = 0b0010_0000;
        const Sprite0Hit = 0b0100_0000;
        const VBlank = 0b1000_0000;
    }
}

impl PpuRegisters {
    
}
